#-------------------------------------------------------------------------------
# Name:        R2P
# Purpose:     Push requirements from excel sheet to ProjeQtor database with
#              all images encapsulated in excel cells and requirements hierarchy
#              v1.3.0
#
# Author:      Laurent CABOT
#
# Created:     13/04/2022
# Copyright:   (c) Laurent CABOT 2022
# Licence:     Free to use, modify, dispatch
#-------------------------------------------------------------------------------
import configparser
import mysql.connector
from mysql.connector import Error
import xml.etree.ElementTree as xml
import zipfile
import os
import sys
import shutil
import openpyxl
import uuid
import progressbar
from datetime import datetime

version = "1.3.0"

# Global variables:
MySQL_Host      =""
MySQL_Port      =""
MySQL_User      =""
MySQL_Password  =""
MySQL_DbName    =""
XLSX            =""
ProjectID       =""
ProductID       =""
VersionID       =""
TypeID          =""
UserID          =""

global cursor
global connection

COLDESC = ""                                                                                                            # Column of the XLSX that contains requirements
ROWMIN = ""
COLEXT = ""
COLNAME = ""                                                                                                            # 1st requirement line in XLSX
PIP=""                                                                                                                  # Projeqtor Image Path

lId=[]                                                                                                                  # List of Ids
lRow=[]                                                                                                                 # List of excel row numbers corresponding to the list of Ids
lOl=[]                                                                                                                  # List of outline levels corresponding to each excel row
lTopId=[]                                                                                                               # List of top Id corresponding to each line

def main():
    global cursor
    def readConfig():
        global MySQL_Host
        global MySQL_Port
        global MySQL_User
        global MySQL_Password
        global MySQL_DbName
        global XLSX
        global ProjectID
        global ProductID
        global VersionID
        global TypeID
        global UserID
        global COLDESC
        global ROWMIN
        global COLEXT
        global COLNAME
        global PIP

        c = configparser.ConfigParser()
        c.read('config.ini')

        try:
            MySQL_Host = c['Database']['MySQL_Host']
        except:
            print ("MySQL_Host not defined in config.ini")
            sys.exit()

        try:
            MySQL_Port = c['Database']['MySQL_Port']
        except:
            print ("MySQL_Port not defined in config.ini")
            sys.exit()

        try:
            MySQL_User = c['Database']['MySQL_User']
        except:
            print ("MySQL_User not defined in config.ini")
            sys.exit()

        try:
            MySQL_Password = c['Database']['MySQL_Password']
        except:
            print ("MySQL_Password not defined in config.ini")
            sys.exit()

        try:
            MySQL_DbName = c['Database']['MySQL_DbName']
        except:
            print ("MySQL_DbName not defined in config.ini")
            sys.exit()

        try:
            XLSX = c['Reqs']['XLSX']
        except:
            print ("XLSX not defined in config.ini")
            sys.exit()

        try:
            COLDESC = int(c['Reqs']['Requirements_description_column'])
        except:
            print ("Requirements_description_column not defined in config.ini")
            sys.exit()

        try:
            ROWMIN = int(c['Reqs']['1st_requirements_row'])
        except:
            print ("1st_requirements_row not defined in config.ini")
            sys.exit()

        try:
            COLEXT = int(c['Reqs']['External_reference_column'])
        except:
            print ("External_reference_column not defined in config.ini")
            sys.exit()

        try:
            COLNAME = int(c['Reqs']['Requirements_name_column'])
        except:
            print ("Requirements_name_column not defined in config.ini")
            sys.exit()

        try:
            ProjectID = c['Projeqtor']['ProjectID']
        except:
            print ("ProjectID not defined in config.ini")
            sys.exit()

        try:
            ProductID = c['Projeqtor']['ProductID']
        except:
            print ("ProductID not defined in config.ini")
            sys.exit()

        try:
            VersionID = c['Projeqtor']['VersionID']
        except:
            print ("VersionID not defined in config.ini")
            sys.exit()

        try:
            TypeID = c['Projeqtor']['TypeID']
        except:
            print ("TypeID not defined in config.ini")
            sys.exit()

        try:
            UserID = c['Projeqtor']['UserID']
        except:
            print ("UserID not defined in config.ini")
            sys.exit()

        try:
            PIP =c['Projeqtor']['Projeqtor_files_images_path']
        except:
            print ("Projeqtor_files_images_path not defined in config.ini")
            sys.exit()
        pass
    def DB_Connect():
        global cursor
        global connection
        try:
            connection = mysql.connector.connect(
                user=MySQL_User,
                password=MySQL_Password,
                host=MySQL_Host,
                database=MySQL_DbName,
                port=MySQL_Port)

            if connection.is_connected():
                db_Info = connection.get_server_info()
                print("Connected to MySQL Server version ", db_Info)
                cursor = connection.cursor()
                cursor.execute("select database();")
                record = cursor.fetchone()
                print("You're connected to database: ", record)
        except Error as e:
            print("Error while connecting to MySQL", e)
        finally:
            cursor = connection.cursor()
        return
    def get_images_from_excel():
        images = []
        document_title, _ = os.path.splitext(os.path.basename(XLSX))
        output_file = document_title + ".reqifz"
        out_zip = zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED)
        try:
            in_excel = zipfile.ZipFile(XLSX)
        except:
            enablePrint()
            print ("File is not openXML compliant. Can't read "+XLSX)
            sys.exit()
        if "xl/drawings/drawing1.xml" not in in_excel.namelist():
            # No image in the excel file
            log("No image found in "+XLSX)
            return (images)

        f = in_excel.open("xl/drawings/drawing1.xml")
        # Parsing the XML file
        try:
            tree = xml.parse(f)
        except:
            print("Error when parsing file: "+f)
            exit()

        root = tree.getroot()

        xdr="{http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing}"
        a="{http://schemas.openxmlformats.org/drawingml/2006/main}"
        r="{http://schemas.openxmlformats.org/officeDocument/2006/relationships}"

        for X0 in root.findall(xdr + 'twoCellAnchor'):
            X1=X0.find(xdr+'from')
            X2=X1.find(xdr+'col')
            col=str(1+int(X2.text))

            X2=X1.find(xdr+'row')
            row=str(1+int(X2.text))

            X1=X0.find(xdr+'pic')                                                                                       # Not grouped pictures
            if(X1):
                X2=X1.find(xdr+'blipFill')
                X3=X2.find(a+'blip')
                ref=X3.attrib["{http://schemas.openxmlformats.org/officeDocument/2006/relationships}embed"]

                images.append({"row": int(row), "col": int(col), "ref" : ref})

            X1=X0.find(xdr+'grpSp')                                                                                     # Grouped picture
            if(X1):
                for X2 in X1.findall(xdr + 'pic'):
                    X3=X2.find(xdr+'blipFill')
                    X4=X3.find(a+'blip')
                    ref=X4.attrib["{http://schemas.openxmlformats.org/officeDocument/2006/relationships}embed"]

                    images.append({"row": int(row), "col": int(col), "ref" : ref})
                    pass
            pass

            f.close()

        f = in_excel.open("xl/drawings/_rels/drawing1.xml.rels")
        # Parsing the XML file
        try:
            tree = xml.parse(f)
        except:
            print("Error when parsing file: "+f)
            exit()

        root = tree.getroot()
        for relation_ship in root:
            id = relation_ship.attrib["Id"]
            target = relation_ship.attrib["Target"]
            for item in images:
                if item["ref"] == id:
                    item["target"] = target.replace("../", "")
                    # copy all images to target:
                    out_zip.writestr(target.replace("../", ""), in_excel.read(target.replace("../","xl/")), zipfile.ZIP_DEFLATED)
        f.close()
        out_zip.close()
        return images
    def getFreeIdWithProject(TableName, MinValue):
        Found=False
        i=MinValue
        L=[]
        while Found==False:
            sql="SELECT id FROM "+TableName+" WHERE id = "+str(i)+" AND idProject = " + str(ProjectID)
            cursor = connection.cursor()
            L=cursor.execute(sql)
            cursor.fetchone()
            if(cursor.rowcount==0):
                return(str(i))
            else:
                i=i+1
        pass
    def getFreeId(TableName, MinValue):
        Found=False
        i=MinValue
        L=[]
        while Found==False:
            sql="SELECT id FROM "+TableName+" WHERE id = "+str(i)
            cursor = connection.cursor()
            L=cursor.execute(sql)
            cursor.fetchone()
            if(cursor.rowcount==0):
                return(str(i))
            else:
                i=i+1
        pass
    def getOutlineLevel(row):
        in_excel = zipfile.ZipFile(XLSX)
        if "xl/worksheets/sheet1.xml" not in in_excel.namelist():
            print("Excel file is not compliant to openxml")
            exit()

        f = in_excel.open("xl/worksheets/sheet1.xml")
        # Parsing the XML file
        try:
            tree = xml.parse(f)
        except:
            print("Error when parsing file: "+f)
            exit()

        root = tree.getroot()
        xmlns1="{http://schemas.openxmlformats.org/spreadsheetml/2006/main}"

        X0=root.find(xmlns1+"sheetData")
        for X1 in X0:
            if (X1.tag ==xmlns1+"row"):
                r=X1.attrib['r']
                if (int(r)==row):
                    try:
                        outlineLevel=X1.attrib['outlineLevel']
                    except:
                        outlineLevel='0'
                    return(outlineLevel)
            pass
        f.close()
    def log(s):
        log = open('log.log','a',encoding="utf-8")
        log.write(s+'\n')
        log.close()
    def blockPrint():
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = open(os.devnull, 'w')
    def enablePrint():
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
    #-------------------------------------------------------------------------------------------------------------------
    global XLSX
    print ("Excel to Projeqtor requirements importer "+version)

    # Remove media folder if it exists and log file:
    try:
        os.remove("log.log")
    except:
        pass
    try:
        shutil.rmtree("media")
    except:
        pass

    readConfig()
    if not (os.path.isfile(XLSX)):
        print("File does not exist: "+XLSX)
        sys.exit()
    if not(os.path.exists(PIP)):
        print("No "+PIP+" in Projeqtor server file system")
        sys.exit()
    if not(os.access(PIP, os.W_OK)):
        print(PIP+" is not writable")
        sys.exit()

    DB_Connect()
    if not (cursor):
        print("Not connected to database")
        print("Process aborted")
    else:
        print("Connected to database at: " + MySQL_Host)

    blockPrint()
    images = get_images_from_excel()
    enablePrint()

    wb = openpyxl.load_workbook(XLSX)
    ws = wb.active

    # Progress bar initialization:
    bar = progressbar.ProgressBar(maxval=(ws.max_row - ROWMIN), \
    widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage()])
    bar.start()
    i=0                                                                                                                 # Progress bar counter

    for row_nr in range(ROWMIN,ws.max_row):
        reqDescription = ws.cell(row_nr,COLDESC).value

        if not reqDescription:
            reqDescription=""

        if (COLNAME != 0):
            reqName = ws.cell(row_nr,COLNAME).value
        else:
            reqName = reqDescription[0:199]

        if (COLEXT!=0):
            reqExternalReference = ws.cell(row_nr,COLEXT).value
        else:
            reqExternalReference = ""

        # Is there any image attached to this cell ?
        for k in range (0, len(images)):
            if ((images[k]['row'] == row_nr) and (images[k]['col']==COLDESC)):
                imageFile = images[k]['target']

                # Extract image file:
                document_title, _ = os.path.splitext(os.path.basename(XLSX))
                output_file = document_title + ".reqifz"
                zipObj=zipfile.ZipFile(output_file, 'r')
                zipObj.extract(imageFile)

                # Rename imageFile to be unique:
                basename = os.path.basename(imageFile)
                extension = basename[-4:]
                dirname = os.path.dirname(imageFile)
                newfilename = uuid.uuid4()

                # Copy the uuid renamed image file to the projeqtor files/images/directory if it does not exist
                while (os.path.isfile(PIP+str(newfilename)+extension)):
                    newfilename = uuid.uuid4()                                                                          # Randomized filename is already present => assign a new one

                try:
                    shutil.copy(imageFile,PIP+str(newfilename)+extension)                                               # The randomized file name is not already present
                except:
                    print("Cannot copy "+imageFile+" image file to: "+PIP)
                    sys.exit()

                # Add an hyperlink to the requirement description and html tags:
                reqDescription = '<div>'+reqDescription + '<br /><img src="../files/images//'+str(newfilename)+extension+'" /><br /><br /></div>'
            pass

        id = getFreeId('requirement',1)

        # Remove media folder:
        try:
            shutil.rmtree("media")
        except:
            pass

        # Manage hierarchy based on excel grouped rows:
        ol=getOutlineLevel(row_nr)                                                                                      # Outline Level
        lId.append(id)
        lRow.append(row_nr)
        lOl.append(ol)
        if (ol=="0"):
            lTopId.append (" ")
        else:
            if (int(lOl[len(lOl)-1]) > int(lOl[len(lOl)-2])):
                lTopId.append (lId[len(lOl)-2])
            if (int(lOl[len(lOl)-1]) == int(lOl[len(lOl)-2])):
                lTopId.append (lTopId[len(lTopId)-1])
            if (int(lOl[len(lOl)-1]) < int(lOl[len(lOl)-2])):
                # Search the same ouline level before:
                j=len(lOl)-2
                while(j>0):
                    if (int(lOl[len(lOl)-1]) == int(lOl[j])):
                        lTopId.append (lTopId[j])
                        break
                    j=j-1
                pass

        # Add creation date and time:
        now = datetime.now()
        dt_string = str(now.strftime("%Y-%m-%d %H:%M:%S"))

        # If name is empty (no name column in XLSX and description cell containing only images), rename:
        if (reqName==""):
            reqName="no name "+str(id)

        # Push to database:
        #                                1  2         3    4           5                 6             7                8         9                      10
        sql = "INSERT INTO requirement ( id,idProject,name,description,externalReference,idRequirement,creationDateTime,idProduct,idTargetProductVersion,idRequirementType ) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"
        args = ([id,ProjectID,reqName,reqDescription,reqExternalReference,lTopId[len(lTopId)-1],dt_string,str(ProductID),str(VersionID),str(TypeID)])
        #        1  2         3       4              5                    6                     7         8              9              10

        cursor = connection.cursor()
        cursor.execute(sql, args)
        connection.commit()

        reqName.replace("\n", "")
        log ("at id="+id+" XLSX line="+str(row_nr)+" outline="+str(ol)+ " ParentID="+  lTopId[len(lTopId)-1] +" inserted: "+reqName[0:64].strip()+"...")
        # Display progress:
        bar.update(i+1)
        i=i+1

    bar.finish()
    print("Finished")
    pass

if __name__ == '__main__':
    main()

    if connection.is_connected():
        cursor.close()
        connection.close()
        print("MySQL connection is closed")
    sys.exit()
